#| -*-Scheme-*-

Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015,
    2016, 2017, 2018, 2019 Matthew Birkholz

This file is part of a planetarium plugin for MIT/GNU Scheme Pucked.

This plugin is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

This plugin is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this plugin; if not, write to the Free Software Foundation,
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

|#

(define-integrable pi/4 (flo:atan2 1. 1.))
(define-integrable pi (flo:* 4. pi/4))
(define-integrable 2pi (flo:* 8. pi/4))
(define-integrable pi/2 (flo:* 2. pi/4))
(define-integrable -pi/2 (flo:negate pi/2))

(define-integrable (flo:2d x y)
  (let ((v (flo:vector-cons 2)))
    (flo:vector-set! v 0 x)
    (flo:vector-set! v 1 y)
    v))
(define-integrable (flo.1 v) (flo:vector-ref v 0))
(define-integrable (flo.2 v) (flo:vector-ref v 1))
(define-integrable (flo.1! v f) (flo:vector-set! v 0 f))
(define-integrable (flo.2! v f) (flo:vector-set! v 1 f))

(define-integrable (flo:3d x y z)
  (let ((v (flo:vector-cons 3)))
    (flo:vector-set! v 0 x)
    (flo:vector-set! v 1 y)
    (flo:vector-set! v 2 z)
    v))
(define-integrable (flo.3 v) (flo:vector-ref v 2))
(define-integrable (flo.3! v f) (flo:vector-set! v 2 f))

(define-integrable (flo:4d r g b a)
  (let ((v (flo:vector-cons 4)))
    (flo:vector-set! v 0 r)
    (flo:vector-set! v 1 g)
    (flo:vector-set! v 2 b)
    (flo:vector-set! v 3 a)
    v))
(define-integrable (flo.4 v) (flo:vector-ref v 3))
(define-integrable (flo.4! v f) (flo:vector-set! v 3 f))

(define-integrable (make-2d-point x y) (flo:2d x y))
(define-integrable (x p) (flo.1 p))
(define-integrable (set-x! p x) (flo.1! p x))
(define-integrable (y p) (flo.2 p))
(define-integrable (set-y! p y) (flo.2! p y))
(define-integrable (make-3d-point x y z) (flo:3d x y z))
(define-integrable (z p) (flo.3 p))
(define-integrable (set-z! p z) (flo.3! p z))
(define-integrable (make-latitude/longitude lat long) (flo:2d lat long))
(define-integrable (latitude p) (flo.1 p))
(define-integrable (set-latitude! p l) (flo.1! p l))
(define-integrable (longitude p) (flo.2 p))
(define-integrable (set-longitude! p l) (flo.2! p l))

(define-integrable (degrees->radians degrees) (flo:* degrees (flo:/ 2pi 360.)))

(define-integrable (radians->degrees radians) (flo:* radians (flo:/ 360. 2pi)))

(define-integrable (dms->degrees degrees minutes seconds)
  (flo:+ (->flonum degrees)
	 (flo:/ (flo:+ (->flonum minutes) (flo:/ (->flonum seconds) 60.)) 60.)))

(define (degree-string degrees)
  (let* ((deg.r (integer-divide (abs degrees) 3600))
	 (min.r (integer-divide (cdr deg.r) 60)))
    (string-append (if (negative? degrees) "-" "+")
		   (number->string (car deg.r) '(int))
		   ":"(string-pad-left (number->string (car min.r) '(int))
				       2 #\0)
		   ":"(string-pad-left (number->string (cdr min.r) '(int))
				       2 #\0))))

(define (lesser-angle a b)
  ;; Angle B translated to (not more than 2pi) numerically less than A.
  (let loop ((b b))
    (cond ((flo:< b (flo:- a 2pi))
	   (loop (flo:+ b 2pi)))
	  ((flo:< b a)
	   b)
	  (else
	   (loop (flo:- b 2pi))))))

(define (greater-angle a b)
  ;; Angle B translated to (not more than 2pi) numerically greater than A.
  (let loop ((b b))
    (cond ((flo:< a (flo:- b 2pi))
	   (loop (flo:- b 2pi)))
	  ((flo:< a b)
	   b)
	  (else
	   (loop (flo:+ b 2pi))))))

(define (canonical-longitude long)
  ;; Equivalent to LONG degrees, in range [0,360).
  (let ((l (flo:truncate long)))
    (let ((frac (flo:- long l))
	  (r (remainder l 360)))
      (let ((r+frac (flo:+ r frac)))
	(if (flo:negative? r)
	    (flo:+ r+frac 360.)
	    r+frac)))))

(define (latitude-longitude->string lat/lng)
  ;; Format a latitude/longitude in the standard form, e.g.
  ;;
  ;;     30°26.400′N 122°44.400′W

  (let ((lat (latitude lat/lng))
	(lng (longitude lat/lng)))
    (let ((neg-lat? (flo:negative? lat))
	  (neg-lng? (flo:negative? lng))
	  (abs-lat (flo:abs lat))
	  (abs-lng (flo:abs lng)))
      (let ((lat-deg (flo:truncate abs-lat))
	    (lng-deg (flo:truncate abs-lng)))
	(let ((lat-min (flo:* (flo:- abs-lat lat-deg) 60.))
	      (lng-min (flo:* (flo:- abs-lng lng-deg) 60.)))
	  (string-append
	   (number->string (inexact->exact lat-deg) '(int))
	   "\xb0;"
	   (number->string lat-min '(fix 3))
	   "\x2032;"
	   (if neg-lat? "S " "N ")
	   (number->string (inexact->exact lng-deg) '(int))
	   "\xb0;"
	   (number->string lng-min '(fix 3))
	   "\x2032;"
	   (if neg-lng? "W" "E")))))))

(define (geodesic-distance p1 p2)
  ;; "[Given] the geographic coordinates of two points on the surface
  ;; of the Earth [...] the shortest distance S between these points,
  ;; measured along the Earth's surface. [...]  We will suppose that
  ;; these points are at sea level [and] no great accuracy is needed."
  ;; -Meeus 2009, specifically the second English edition (1998) "with
  ;; corrections as of August 10, 2009" of _Astronomical_Algorithms_
  ;; by Jean Meeus.

  (let ((d (angular-separation p1 p2)))
    (flo:* d (flo:/ (flo:* 6371000.0 pi)
		    180.0))))

(define (angular-separation p1 p2)
  ;; "Mr. Thierry Pauwels, of the Royal Observatory of Belgium,
  ;; communicated the following method. [...]  Mathematically
  ;; speaking, this method is completely identical to formula 17.1,
  ;; but a computer will yield more accurate results from an
  ;; arctangent than from an arccosine." [p.115] -Meeus 2009

  (let ((lat1 (degrees->radians (latitude p1)))
	(lng1 (degrees->radians (longitude p1)))
	(lat2 (degrees->radians (latitude p2)))
	(lng2 (degrees->radians (longitude p2))))
    (let ((x (flo:- (flo:* (flo:cos lat1) (flo:sin lat2))
		    (flo:* (flo:sin lat1)
			   (flo:* (flo:cos lat2)
				  (flo:cos (flo:- lng2 lng1))))))
	  (y (flo:* (flo:cos lat2) (flo:sin (flo:- lng2 lng1))))
	  (z (flo:+ (flo:* (flo:sin lat1) (flo:sin lat2))
		    (flo:* (flo:cos lat1)
			   (flo:* (flo:cos lat2)
				  (flo:cos (flo:- lng2 lng1)))))))
      (let ((d (flo:atan (flo:/ (flo:sqrt (flo:+ (flo:* x x) (flo:* y y)))
				z))))
	(radians->degrees
	 (if (flo:negative? z)
	     (flo:+ pi d)
	     d))))))

(define (test-angular-separation)

  (define (test lat1 lng1 lat2 lng2 degrees tolerance)
    (let ((d (angular-separation
	      (make-latitude/longitude lat1 lng1)
	      (make-latitude/longitude lat2 lng2))))
      (if (not (flo:~= degrees d tolerance))
	  (warn "Wrong angular separation:" d degrees))))

  ;; Arcturus to Spica
  (test 19.1825 213.9154		;Arcturus
	-11.1614 201.2983		;Spica
	32.7930 0.00005)

  ;; Front to nearly the back.
  (test 0. 0.
	0. 135.
	135. 0.0000000000005)

  ;; Across the N pole.
  (test 80. 20.
	80. 200.
	20. 0.0000000000005)

  ;; Top to nearly bottom on right.
  (test 90. 90.
	-85. 90.
	175. 0.0000000000005))

(define (2d-translate! p dx dy q)
  (let ((x (x p))
	(y (y p)))
    (flo:vector-set! q 0 (flo:+ x dx))
    (flo:vector-set! q 1 (flo:+ y dy))))

(define (2d-rotate! p angle q)
  (let ((cos (flo:cos angle))
	(sin (flo:sin angle)))
    (let ((x (x p))
	  (y (y p)))
      (flo:vector-set! q 0 (flo:+ (flo:* x cos) (flo:* y (flo:negate sin))))
      (flo:vector-set! q 1 (flo:+ (flo:* x sin) (flo:* y cos))))))

(define (flo:~= n1 n2 epsilon)
  ;; aka approximately-=
  (flo:< (flo:abs (flo:- n2 n1)) epsilon))

(define (2d-~= p1 p2 epsilon)
  (and (flo:~= (x p1) (x p2) epsilon)
       (flo:~= (y p1) (y p2) epsilon)))

(define (3d-~= p1 p2 epsilon)
  (and (flo:~= (x p1) (x p2) epsilon)
       (flo:~= (y p1) (y p2) epsilon)
       (flo:~= (z p1) (z p2) epsilon)))