#| -*-Scheme-*-

Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015,
    2016, 2017, 2018, 2019 Matthew Birkholz

This file is part of a planetarium plugin for MIT/GNU Scheme Pucked.

This plugin is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

This plugin is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this plugin; if not, write to the Free Software Foundation,
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

|#

(define (make-tellurion)

  (define queue (make-thread-queue))

  (define (queue! command . args)
    (thread-queue/queue! queue (cons command args))
    unspecific)

  (define (tellurion command . args)
    (cond ((eq? command 'turn-to)
	   (let ((lat (->flonum (car args)))
		 (long (->flonum (cadr args))))
	     (if (or (< lat -90) (< 90 lat))
		 (error "Invalid latitude:" lat))
	     (if (or (< long -180) (< 180 long))
		 (error "Invalid longitude:" long))
	     (queue! 'turn-to lat long)))
	  ((eq? command 'time-to)
	   (queue! 'time-to (->time (car args))))
	  ((eq? command 'hour-to)
	   (let ((hour (car args)))
	     (guarantee integer? hour 'hour-to)
	     (if (or (< hour 0) (< 23 hour))
		 (error "Invalid hour:" hour))
	     (queue! 'hour-to hour)))
	  ((eq? command 'stop)
	   (queue! queue 'stop))
	  ((eq? command 'go)
	   (queue! queue 'go))
	  (else (error "unknown command:" command args))))

  (detach-thread (create-thread #f
		   (lambda () (run-tellurion queue))
		   tellurion))
  tellurion)

(define (->time object)
  (cond ((eq? object 'current)
	 'current)
	((decoded-time? object)
	 (decoded-time->universal-time object))
	((integer? object)
	 object)
	(else
	 (error "Not a time (universal, decoded or 'current):" object))))

(define (run-tellurion queue)
  (let ((device (make-suitable-graphics-device))
	;; Latitude and longitude per GPS -- positive degrees long. to the east.
	(lat/lng (make-latitude/longitude 23.271 0.))
	(time 'current)
	(stopped? #f))
    (define (draw)
      (draw-tellurion device
		      (if (eq? time 'current)
			  (get-universal-time)
			  time)
		      lat/lng))
    (draw)
    (let loop ()
      (let ((command (if stopped?
			 (thread-queue/dequeue! queue)
			 (thread-queue/dequeue-no-hang!
			  queue
			  (if (eq? time 'current)
			      (let* ((time (get-universal-time))
				     (next (* 15 60
					      (1+ (quotient time (* 15 60)))))
				     (sec (- next time)))
				(* sec 1000))
			      (* 15 60 1000))))))
	(cond ((eq? command #f)	;timeout
	       (if (not (eq? time 'current))
		   (set! time (+ time (* 15 60))))
	       (draw))
	      ((eq? (car command) 'turn-to)
	       (set-latitude! lat/lng (cadr command))
	       (set-longitude! lat/lng (caddr command))
	       (draw))
	      ((eq? (car command) 'time-to)
	       (set! time (cadr command))
	       (draw))
	      ((eq? (car command) 'stop)
	       (set! stopped? #t))
	      ((eq? (car command) 'go)
	       (set! stopped? #f)
	       (if (eq? time 'current)
		   (draw)))))
      (loop))))

(define draw-tellurion
  (let ((lat/lng-pos (make-2d-point -1. -1.02))
	(time-pos (make-2d-point -1. 1.))
	(cross-hair-left (make-2d-point -.02 0.))
	(cross-hair-right (make-2d-point .02 0.))
	(cross-hair-top (make-2d-point 0. .02))
	(cross-hair-bottom (make-2d-point 0. -.02)))
    (named-lambda (draw-tellurion device time lat/lng)
      (clear-graphics device)
      (draw-text device time-pos (universal-time->string time) "black")
      (draw-text device lat/lng-pos (latitude-longitude->string lat/lng) "black")
      (draw-earth device
		  (orientation-matrix lat/lng)
		  (solar-latitude/longitude
		   (universal-time->julian-day time)))
      (draw-segment device cross-hair-left cross-hair-right "black")
      (draw-segment device cross-hair-top cross-hair-bottom "black")
      (flush-graphics device))))

(define (orientation-matrix lat/lng)
  (let ((Mx (make-x-rotation-matrix (degrees->radians
				     (latitude lat/lng))))
	(My (make-y-rotation-matrix (degrees->radians
				     (flo:negate (longitude lat/lng))))))
    (3d-multiply! Mx My Mx)
    Mx))