#| -*-Scheme-*-

Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015,
    2016, 2017, 2018, 2019 Matthew Birkholz

This file is part of a planetarium plugin for MIT/GNU Scheme Pucked.

This plugin is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

This plugin is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this plugin; if not, write to the Free Software Foundation,
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

|#

;;;; Draw Elevation data.

(declare (integrate-external "geometry"))

(define (make-terrain . options)
  (let ((widget (%make-terrain options)))
    (named-lambda (terrain command . args)

      (define (check-flonum-args num description)
	(if (not (= num (guarantee-list-of-type->length
			 args real? "a real number" 'terrain)))
	    (error (string-append
		    "The "(symbol-name command)" command requires "
		    (number->string num '(int))" arguments: "description"."))))

      (cond
       ((eq? 'position command)
	(check-flonum-args 3 "latitude, longitude and altitude")
	(let ((lat (->flonum (car args)))
	      (lng (->flonum (cadr args)))
	      (alt (->flonum (caddr args))))
	  (if (not (flo:<= (flo:abs lat) 80.))
	      (error "Requested latitude is greater than 80°:" lat))
	  (let ((pos (glx-viewport-position widget))
		(origin (terrain-viewport-origin widget))
		(m/d (flo:/ (terrain-viewport-step-meters widget)
			    (terrain-viewport-step-degrees widget))))
	    (set-x! pos (flo:* m/d (flo:- lng (longitude origin))))
	    (set-y! pos alt)
	    (set-z! pos (flo:* m/d (flo:- (latitude origin) lat)))))
	(gtk-widget-queue-draw widget))

       ((eq? 'heading command)
	(check-flonum-args 1 "azimuth")
	(set-glx-viewport-heading! widget
				   (degrees->radians (->flonum (car args))))
	(gtk-widget-queue-draw widget))

       ((eq? 'tilt command)
	(check-flonum-args 1 "inclination in degrees")
	(if (not (flo:<= (flo:abs (->flonum (car args))) 80.))
	    (error "Requested inclination greater than 80°.")
	    (set-glx-viewport-tilt! widget (degrees->radians (car args))))
	(gtk-widget-queue-draw widget))

       (else
	(error:wrong-type-argument command "a command name" 'terrain))))))

(define (%make-terrain options)

  (define (get-keyword-value klist key default)
    (let loop ((klist klist))
      (if (pair? klist)
	  (if (eq? (car klist) key)
	      (car (cdr klist))
	      (loop (cdr (cdr klist))))
	  default)))

  (define no-value "<none>")

  (define (get-flonum-option name)
    (let ((v (get-keyword-value options name no-value)))
      (guarantee real? v name)
      (->flonum v)))

  (define (get-integer-option name default)
    (let ((v (get-keyword-value options name no-value)))
      (if (eq? v no-value)
	  default
	  (begin
	    (guarantee integer? v name)
	    v))))

  (let ((filename (get-keyword-value options 'filename #f))
	(rows (get-integer-option 'rows #f))
	(columns (get-integer-option 'columns #f))
	(latitude (get-flonum-option 'latitude))
	(longitude (get-flonum-option 'longitude))
	(step (get-flonum-option 'step))
	(color-function
	 (let ((v (get-keyword-value options 'color-function #f)))
	   (and v (begin
		    (guarantee-procedure-of-arity
		     v 3 'make-terrain:color-function)
		    v))))
	(width (get-integer-option 'width 300))
	(height (get-integer-option 'height 300)))
    (let ((matrix (make-heightmap-matrix filename rows columns))
	  (origin (make-latitude/longitude latitude longitude))
	  (title (if filename
		     (string-append "Terrain: "(->namestring filename))
		     "Terrain")))
      (let ((window (gtk-window-new 'toplevel))
	    (grid (gtk-grid-new))
	    (label (gtk-label-new "")))
	(gtk-window-set-title window title)
	(gtk-container-set-border-width window 5)
	(gtk-widget-set-hexpand label #t)
	(let* ((min.max (matrix-min.max matrix rows columns))
	       (widget (make-terrain-viewport
			matrix rows columns (car min.max) (cdr min.max)
			origin step color-function
			label width height)))
	  (gtk-widget-set-vexpand widget #t)
	  (gtk-widget-set-hexpand widget #t)
	  (gtk-grid-attach grid widget 0 0 1 1)
	  (gtk-grid-attach grid label 0 1 1 1)
	  (gtk-container-add window grid)
	  (gtk-widget-show-all window)
	  widget)))))

(define-class (<terrain-viewport>
	       (constructor make-terrain-viewport
			    (matrix rows columns min-height max-height
			     origin step-degrees color-function label)
			    (width height)))
    (<glx-viewport>)

  ;; Matrix should contain metric height information per latitude/
  ;; longitude.  MATRIX[0,0] would be the height in meters at origin,
  ;; a latitude/longitude.  MATRIX[0,1] is the height at a position
  ;; step-degrees due East.  MATRIX[1,0] is the height step-degrees
  ;; North.
  (matrix define accessor)
  (rows define accessor)
  (columns define accessor)
  (min-height define accessor)
  (max-height define accessor)

  (origin define accessor)
  (step-degrees define accessor)
  (color-function define accessor)

  ;; A GtkLabel to update with the current lat/lng, heading, and inclination.
  (label define accessor)

  ;; We scale the matrix indices by step-meters -- meters along the
  ;; earth's surface for step-degrees of longitude (at the origin's
  ;; latitude).  No fancy project is used.  It is assumed the
  ;; latitudes are small and/or do not vary greatly.
  (step-meters define standard)

  ;; Update light0 if changing this:
  (light-position define accessor
		  initializer (lambda () (flo:4d 1. 1. 1. 0.)))

  ;; The display-list for the terrain mesh.
  (mesh define standard initial-value #f))

(define-method initialize-instance ((widget <terrain-viewport>) width height)
  (%trace "; (initialize-instance <terrain-viewport>)\n")
  (call-next-method widget width height)
  (let ((origin (terrain-viewport-origin widget))
	(step-degrees (terrain-viewport-step-degrees widget))
	(columns (->flonum (terrain-viewport-columns widget))))
    (let ((lat (latitude origin))
	  (lng-orig (longitude origin)))
      (let ((lng-end (flo:+ lng-orig (flo:* step-degrees columns))))
	(let ((step-meters
	       (flo:/ (geodesic-distance
		       (make-latitude/longitude lat lng-orig)
		       (make-latitude/longitude lat lng-end))
		      columns)))
	  (set-terrain-viewport-step-meters! widget step-meters)
	  (set-glx-viewport-step-size! widget 100.)
	  (set-glx-viewport-near! widget 10.)
	  (set-glx-viewport-far! widget 150000.))))))

(define-method fix-widget-realize-callback ((widget <terrain-viewport>))
  (%trace "; (fix-widget-realize-callback <terrain-viewport>)\n")
  (call-next-method widget)
  (with-glx-widget widget
    (lambda ()
      (gl:shade-model 'smooth)		   ; Enable Smooth Shading
      ;;(gl:shade-model 'flat)
      (gl:clear-color (color .527 .805 .977 1.)) ; light sky blue: #x87cefa
      (gl:enable 'depth-test)
      ;;(gl:enable 'cull-face)
      ;; Really Nice Perspective Calculations
      (gl:hint 'perspective-correction 'nicest)

      (gl:light-model 'local-viewer 1.)
      (gl:enable 'lighting)
      (gl:enable 'light0)
      (if (terrain-viewport-color-function widget)
	  (begin
	    (gl:color-material 'front 'diffuse)
	    (gl:enable 'color-material)))

      (let ((mesh (gl:gen-lists 1)))
	(gl:new-list mesh 'compile)
	(draw-mesh widget)
	(gl:end-list)
	(set-terrain-viewport-mesh! widget mesh))))
  #t)

(define-method glx-viewport-draw ((widget <terrain-viewport>))
  (%trace2 "; (glx-viewport-draw <terrain-viewport>)\n")
  (gl:light 'light0 'position (terrain-viewport-light-position widget))
  (gl:call-list (terrain-viewport-mesh widget))
  (update-label widget))

(define (update-label widget)
  (%trace2 "; update-label\n")
  (let ((origin (terrain-viewport-origin widget))
	(pos (glx-viewport-position widget))
	(d/m (flo:/ (terrain-viewport-step-degrees widget)
		    (terrain-viewport-step-meters widget))))

    (define (%_5d i)
      (let ((s (number->string i '(int))))
	(if (< (string-length s) 5)
	    (string-pad-left s 5 #\space)
	    s)))

    (define (meters->degrees m) (flo:* m d/m))

    (gtk-label-set-text
     (terrain-viewport-label widget)
     (string-append
      (latitude-longitude->string
       (make-latitude/longitude
	(flo:- (latitude origin) (meters->degrees (z pos)))
	(flo:+ (longitude origin) (meters->degrees (x pos)))))
      "  alt: "(%_5d (flo:round->exact (y pos)))
      "m  head: "(number->string
		  (flo:round->exact
		   (radians->degrees (glx-viewport-heading widget))) '(int))
      "\xb0; incl: "(number->string
		  (flo:round->exact
		   (radians->degrees (glx-viewport-tilt widget))) '(int))
      "\xb0;"))))

(define (draw-sea-level widget)
  ;; A translucent blue plane at sea level.
  (%trace "; draw-sea-level\n")
  (let ((rows (terrain-viewport-rows widget))
	(columns (terrain-viewport-columns widget))
	(step (terrain-viewport-step-meters widget)))
    (%trace "; disabling lighting\n")
    (gl:disable 'lighting)
    (%trace "; enabling blend\n")
    (gl:enable 'blend)
    (%trace "; setting blend-func\n")
    (gl:blend-func 'src-alpha 'one-minus-src-alpha)
    (%trace "; setting color\n")
    (gl:color (flo:4d 0. 0. 1. .1))
    (%trace "; setting normal\n")
    (gl:normal (flo:3d 0. 1. 0.))
    (%trace "; beginning quads\n")
    (gl:begin 'quads)
    (let ((max-x (flo:* (flo:- (->flonum columns) 1.) step))
	  (min-z (flo:* (flo:- 1. (->flonum rows)) step)))
      (gl:vertex (flo:3d    0. 0.    0.))
      (gl:vertex (flo:3d max-x 0.    0.))
      (gl:vertex (flo:3d max-x 0. min-z))
      (gl:vertex (flo:3d    0. 0. min-z)))
    (%trace "; ending quads\n")
    (gl:end)
    (%trace "; disabling blend\n")
    (gl:disable 'blend)
    (%trace "; enabling lighting\n")
    (gl:enable 'lighting)))

(define (draw-mesh widget)
  (%trace "; draw-mesh "widget"\n")
  (let ((matrix (terrain-viewport-matrix widget))
	(rows (terrain-viewport-rows widget))
	(columns (terrain-viewport-columns widget))
	(step (terrain-viewport-step-meters widget))
	(-step (flo:- 0. (terrain-viewport-step-meters widget)))
	(color-function (terrain-viewport-color-function widget))
	(min-height (terrain-viewport-min-height widget))
	(max-height (terrain-viewport-max-height widget)))

    (define (vertex x y)
      (let ((h (matrix-ref matrix columns x y)))

	(define (ref x y)
	  (if (or (fix:= x -1)
		  (fix:= x rows)
		  (fix:= y -1)
		  (fix:= y columns))
	      #f
	      (matrix-ref matrix columns x y)))

	(if color-function
	    (gl:color (color-function h min-height max-height)))
	(gl:normal (normal step h
			   (ref (fix:1+ x) y)	 ; east
			   (ref (fix:-1+ x) y)	 ; west
			   (ref x (fix:1+ y))	 ; north
			   (ref x (fix:-1+ y)))) ; south
	(gl:vertex (flo:3d (flo:* (->flonum x) step)
			   h
			   (flo:* (->flonum y) -step)))))

    (%trace ";   begin quads\n")
    (gl:begin 'quads)			; or 'lines for wireframe
    (let ((last-row (fix:- rows 1))
	  (last-column (fix:- columns 1)))
      (do ((y 0 (fix:1+ y)))
	  ((fix:= y last-row))
	(do ((x 0 (fix:1+ x)))
	    ((fix:= x last-column))
	  (vertex          x          y)
	  (vertex (fix:1+ x)          y)
	  (vertex (fix:1+ x) (fix:1+ y))
	  (vertex          x (fix:1+ y)))))
    (gl:end)
    (%trace ";   end quads\n")
    ;;(draw-sea-level widget)
    ))

(define (normal step height east west north south)
  (let ((-step (flo:- 0. step))
	(east (if (not east) 0. (flo:- east height)))
	(west (if (not west) 0. (flo:- west height)))
	(north (if (not north) 0. (flo:- north height)))
	(south (if (not south) 0. (flo:- south height))))
    (let ((n (3d-sum
	      (3d-sum
	       (normalized-3d-cross-product
		(flo:3d   step east      0.)
		(flo:3d     0. north  -step))
	       (normalized-3d-cross-product
		(flo:3d     0. north  -step)
		(flo:3d  -step west      0.)))
	      (3d-sum
	       (normalized-3d-cross-product
		(flo:3d  -step west      0.)
		(flo:3d     0. south   step))
	       (normalized-3d-cross-product
		(flo:3d     0. south   step)	
		(flo:3d   step east      0.))))))
      (normalize-3d! n)
      n)))

(define (height-color height min max)
  (if (flo:negative? height)
      (let ((-norm (flo:- 1. (flo:/ height min))))
	(let #;((r (flo:+ .1 (flo:* -norm .8)))
	      (g (flo:+ .1 (flo:* -norm .8)))
	      (b (flo:+ .4 (flo:* -norm .6))))
	    ((r (flo:+ .2 (flo:* -norm .6)))
	     (g (flo:+ .2 (flo:* -norm .6)))
	     (b (flo:+ .2 (flo:* -norm .8))))
	  (color r g b 1.)))
      (let ((norm (flo:/ height max)))
	(let ((r norm)
	      (g 1.)
	      (b 0. #;(flo:* norm .4)))
	  (color r g b 1.)))))

(define-integrable (normalized-3d-cross-product v1 v2)
  (let ((p (3d-cross-product v1 v2)))
    (normalize-3d! p)
    p))

(define (test-normals)
  (let ((n 1))

    (define (test correct computed)
      #;(for-each display
		`("; Normal test ",n":\n"
		  "computed:",computed"\n"
		  " correct:",correct"\n"))
      (if (not (3d-~= computed correct 5.e-16))
	  (warn "; Normal test failed:" n computed correct))
      (set! n (1+ n)))

    (test (flo:3d 0. 1. 0.)
	  (normal 1. 0. 1. 1. 1. 1.))
    (test (flo:3d 0. 1. 0.)
	  (normal 1. 1. 0. 0. 0. 0.))
    (test (flo:3d (- (cos pi/4)) (sin pi/4) 0.)
	  (normal 1. 0. 1. -1. 0. 0.))
    (test (flo:3d (cos pi/4) (sin pi/4) 0.)
	  (normal 1. 0. -1. 1. 0. 0.))
    (test (flo:3d 0. (sin pi/4) (- (cos pi/4)))
	  (normal 1. 0. 0. 0. -1. 1.))
    (test (flo:3d 0. (sin pi/4) (- (cos pi/4)))
	  (normal 1. 0. 1. 1. -1. 1.))
    (test (flo:3d 0. (sin pi/4) (- (cos pi/4)))
	  (normal 1. 0. -1. -1. -1. 1.))
    (test (flo:3d 0. (sin pi/4) (cos pi/4))
	  (normal 1. 0. 0. 0. 1. -1.))
    (test (flo:3d 0. (sin pi/4) (cos pi/4))
	  (normal 1. 0. 1. 1. 1. -1.))
    (test (flo:3d 0. (sin pi/4) (cos pi/4))
	  (normal 1. 0. -1. -1. 1. -1.))
    ))

(define-integrable-operator (3d-sum a b)
  (flo:3d (flo:+ (x a) (x b))
	  (flo:+ (y a) (y b))
	  (flo:+ (z a) (z b))))

(define-integrable-operator (2d-sum a b)
  (flo:2d (flo:+ (x a) (x b))
	  (flo:+ (y a) (y b))))

(define-integrable-operator (normalize-3d! v)
  (let ((l (sqrt (flo:+
		  (flo:* (x v) (x v))
		  (flo:+
		   (flo:* (y v) (y v))
		   (flo:* (z v) (z v)))))))
    (set-x! v (flo:/ (x v) l))
    (set-y! v (flo:/ (y v) l))
    (set-z! v (flo:/ (z v) l))))

(define-integrable-operator (normalize-2d! v)
  (let ((l (sqrt (flo:+ (flo:* (x v) (x v))
			(flo:* (y v) (y v))))))
    (set-x! v (flo:/ (x v) l))
    (set-y! v (flo:/ (y v) l))))

(define (3d-cross-product v1 v2)
  (flo:3d
   (flo:- (flo:* (y v1) (z v2)) (flo:* (z v1) (y v2)))
   (flo:- (flo:* (z v1) (x v2)) (flo:* (x v1) (z v2)))
   (flo:- (flo:* (x v1) (y v2)) (flo:* (y v1) (x v2)))))

(define (make-heightmap-matrix filename rows columns)
  (let ((matrix (flo:vector-cons (fix:* rows columns))))
    (call-with-input-file filename
      (lambda (in)
	(do ((y 0 (fix:1+ y)))
	    ((fix:= y rows))
	  (do ((x 0 (fix:1+ x)))
	      ((fix:= x columns))
	    (let ((line (read-line in)))
	      (if (eof-object? line) (error "No more data?"))
	      (let ((fields (burst-string line #\tab #f)))
		(let ((height (->flonum
			       (string->number (list-ref fields 2) 'I 'D))))
		  (matrix-set! matrix columns x y height))))))))
    matrix))

(define (write-function-terrain filename
				x-min x-max x-samples
				y-min y-max y-samples
				function)
  (with-output-to-file filename
    (lambda ()
      (let ((time-string
	     (universal-time->local-time-string (get-universal-time))))
	(let ((x-step (/ (- x-max x-min) x-samples)) 
	      (y-step (/ (- y-max y-min) y-samples)))
	  (do ((x x-min (+ x x-step))
	       (n 0 (fix:1+ n)))
	      ((fix:= n x-samples))
	    (do ((y y-min (+ y y-step))
		 (n 0 (fix:1+ n)))
		((fix:= n y-samples))
	      (write-string time-string)
	      (write-char #\tab)
	      (write-string (number->string x '(fix 3)))
	      (write-char #\,)
	      (write-string (number->string y '(fix 3)))
	      (write-char #\tab)
	      (write-string (number->string (function x y) '(fix 3)))
	      (write-char #\newline))))))))

(define-integrable (matrix-set! m c x y v)
  (flo:vector-set! m (fix:+ x (fix:* c y)) v))

(define-integrable (matrix-ref m c x y)
  (flo:vector-ref m (fix:+ x (fix:* c y))))

(define (matrix-min.max matrix rows columns)
  (let ((min #f)
	(max #f))
    (do ((y 0 (fix:1+ y)))
	((fix:= y rows))
      (do ((x 0 (fix:1+ x)))
	  ((fix:= x columns))
	(let ((h (matrix-ref matrix columns x y)))
	  (if (or (not min)
		  (flo:< h min))
	      (set! min h))
	  (if (or (not max)
		  (flo:< max h))
	      (set! max h)))))
    (cons min max)))

(define-integrable color flo:4d)

(define-integrable-operator (flo:max a b)
  (if (flo:< a b) b a))

(define %trace? #f)

(define-syntax %trace
  (syntax-rules ()
    ((_ . msg)
     (if %trace? ((lambda () (outf-error . msg)))))))

(define %trace2? #f)

(define-syntax %trace2
  (syntax-rules ()
    ((_ . msg)
     (if %trace2? ((lambda () (outf-error . msg)))))))